import logging
from pathlib import Path
from time import sleep
from typing import List

import kubernetes as k8s

from beiboot.configuration import ClientConfiguration, default_configuration
from beiboot.utils import (
    create_beiboot_custom_ressource,
    decode_kubeconfig,
    save_kubeconfig_to_file,
    delete_kubeconfig_file,
)

from beiboot.utils import (
    start_kubeapi_portforwarding,
    kill_kubeapi_portforwarding,
)

logger = logging.getLogger("getdeck.beiboot")


def create_cluster(
    cluster_name: str,
    ports: List[str] = None,
    connect: bool = True,
    configuration: ClientConfiguration = default_configuration,
) -> None:
    #
    # 1. create the CR object for Beiboot; apply it
    #
    logger.info(f"Now creating Beiboot {cluster_name}")
    if ports is None:
        ports = []
    obj = create_beiboot_custom_ressource(configuration, cluster_name, ports)
    try:
        configuration.K8S_CUSTOM_OBJECT_API.create_namespaced_custom_object(
            namespace=configuration.NAMESPACE,
            body=obj,
            group="getdeck.dev",
            plural="beiboots",
            version="v1",
        )
    except k8s.client.exceptions.ApiException as e:
        if e.status == 404:
            raise RuntimeError(
                "This cluster does probably not support Getdeck Beiboot, or is not ready."
            )
        elif e.status == 409:
            # this cluster already exists
            raise RuntimeError(
                f"The requested cluster name {cluster_name} already exists."
            )
        else:
            # TODO handle that case
            raise

    if connect:
        logger.info(
            "Now connecting to the cluster; this may take a while to complete. "
        )
        establish_connection(cluster_name, configuration)


def remove_cluster(
    cluster_name: str, configuration: ClientConfiguration = default_configuration
) -> None:
    logger.info(f"Now removing Beiboot {cluster_name}")
    try:
        configuration.K8S_CUSTOM_OBJECT_API.delete_namespaced_custom_object(
            namespace=configuration.NAMESPACE,
            name=cluster_name,
            group="getdeck.dev",
            plural="beiboots",
            version="v1",
        )
    except k8s.client.exceptions.ApiException as e:
        logger.error(f"Error deleting Beiboot object: {e.reason} ({e.status})")
        if e.status == 404:
            #  Getdeck Beiboot probably not available
            pass
        else:
            # TODO  handle that case
            raise
    delete_kubeconfig_file(configuration, cluster_name)
    kill_kubeapi_portforwarding(configuration, cluster_name)


def get_connection(
    cluster_name: str, configuration: ClientConfiguration = default_configuration
) -> str:
    #
    # 1. Wait for the kubeconfig
    #
    i = 0
    while configuration.CLUSTER_CREATION_TIMEOUT > i:
        try:
            bbt = configuration.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
                group="getdeck.dev",
                version="v1",
                namespace=configuration.NAMESPACE,
                plural="beiboots",
                name=cluster_name,
            )
        except k8s.client.exceptions.ApiException as e:
            if e.status == 404:
                raise RuntimeError("This cluster name does not exist")
            else:
                raise RuntimeError(f"Error fetching the Beiboot object: {e.reason}")
        if bbt.get("kubeconfig"):
            # if the kubeconfig was added, this cluster is ready
            break
        else:
            logger.debug(
                f"Waiting for the cluster to become ready {i}/{configuration.CLUSTER_CREATION_TIMEOUT}"
            )
            i = i + 1
            sleep(1)
    #
    # 2. store the kubeconfig to a well-known place
    #
    kubeconfig_object = bbt.get("kubeconfig")  # noqa
    kubeconfig = decode_kubeconfig(kubeconfig_object)
    kubeconfig_file = save_kubeconfig_to_file(configuration, cluster_name, kubeconfig)
    logger.info(f"KUBECONFIG file for {cluster_name} written to: {kubeconfig_file}.")
    return kubeconfig_file


def establish_connection(
    cluster_name: str, configuration: ClientConfiguration = default_configuration
) -> None:

    kubeconfig_location = get_connection(cluster_name, configuration)
    kubeconfig = Path(kubeconfig_location)
    if not kubeconfig.is_file():
        raise RuntimeError(
            f"KUBECONFIG for cluster {cluster_name} not found. Cannot establish a connection."
        )
    start_kubeapi_portforwarding(configuration, cluster_name)
    logger.info(
        f"You can now set 'export KUBECONFIG={kubeconfig_location}' and work with the cluster."
    )


def terminate_connection(
    cluster_name: str, configuration: ClientConfiguration = default_configuration
) -> None:
    kill_kubeapi_portforwarding(configuration, cluster_name)



import sys
import logging
from pathlib import Path

console = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter("[%(levelname)s] %(message)s")
console.setFormatter(formatter)

logger = logging.getLogger("getdeck.beiboot")
logger.addHandler(console)

__VERSION__ = "0.5.0"


class ClientConfiguration(object):
    def __init__(
        self,
        docker_client=None,
        namespace: str = "getdeck",
        registry_url: str = None,
        tooler_image: str = None,
        cluster_timeout: int = 60,
        api_port: int = 6443,
        kube_config_file: str = None,
        kube_context: str = None,
    ):
        self.NAMESPACE = namespace
        self.REGISTRY_URL = (
            registry_url.rstrip("/") if registry_url else "quay.io/getdeck"
        )
        if registry_url:
            logger.debug(
                f"Using registry prefix (other than default): {self.REGISTRY_URL}"
            )
        self.TOOLER_IMAGE = (
            # tooler_image or f"{self.REGISTRY_URL}/tooler:{__VERSION__}"
            tooler_image
            or f"{self.REGISTRY_URL}/tooler:latest"
        )
        if tooler_image:
            logger.debug(f"Using Tooler image (other than default): {tooler_image}")

        if docker_client:
            self.DOCKER = docker_client

        self.KUBECONFIG_FILE = kube_config_file
        self.CLUSTER_CREATION_TIMEOUT = (
            cluster_timeout  # cluster timeout for the kubeconfig
        )
        self.KUBECONFIG_LOCATION = Path.home().joinpath(".getdeck")
        self.KUBECONFIG_LOCATION.mkdir(parents=True, exist_ok=True)
        self.BEIBOOT_API_PORT = api_port
        self.context = kube_context

    def _init_docker(self):
        import docker

        try:
            self.DOCKER = docker.from_env()
        except docker.errors.DockerException as de:
            logger.fatal(f"Docker init error: {de}")
            raise docker.errors.DockerException(
                "Docker init error. Docker host not running?"
            )

    def _init_kubeapi(self, context=None):
        from kubernetes.client import (
            CoreV1Api,
            RbacAuthorizationV1Api,
            AppsV1Api,
            CustomObjectsApi,
        )
        from kubernetes.config import load_kube_config

        if self.KUBECONFIG_FILE:
            load_kube_config(self.KUBECONFIG_FILE, context=context or self.context)
        else:
            load_kube_config(context=context or self.context)
        self.K8S_CORE_API = CoreV1Api()
        self.K8S_RBAC_API = RbacAuthorizationV1Api()
        self.K8S_APP_API = AppsV1Api()
        self.K8S_CUSTOM_OBJECT_API = CustomObjectsApi()

    def __getattr__(self, item):
        if item in [
            "K8S_CORE_API",
            "K8S_RBAC_API",
            "K8S_APP_API",
            "K8S_CUSTOM_OBJECT_API",
        ]:
            try:
                return self.__getattribute__(item)
            except AttributeError:
                self._init_kubeapi()
        if item == "DOCKER":
            try:
                return self.__getattribute__(item)
            except AttributeError:
                self._init_docker()

        return self.__getattribute__(item)

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if k.isupper()}

    def __str__(self):
        return str(self.to_dict())


default_configuration = ClientConfiguration()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beiboot']

package_data = \
{'': ['*']}

install_requires = \
['docker>=5.0.3,<6.0.0', 'kubernetes>=23.3.0,<24.0.0']

entry_points = \
{'console_scripts': ['create = testing.exec:create_a_cluster',
                     'remove = testing.exec:remove_a_cluster']}

setup_kwargs = {
    'name': 'beiboot',
    'version': '0.5.0',
    'description': 'Getdeck Beiboot client project.',
    'long_description': '<p align="center">\n  <img src="https://github.com/Getdeck/beiboot/raw/main/docs/static/img/logo.png" alt="Gefyra Logo"/>\n</p>\n\n# Getdeck Beiboot\n_Getdeck Beiboot_ (or just Beiboot for brevity) is a Kubernetes-in-Kubernetes solution. It was born from the idea to \nprovide Getdeck users a simple yet flexible solution to spin up **hybrid cloud development** infrastructure. This is \nuseful for Kubernetes development workloads which grew too large to run on a development machine (or with pony workloads \non MacOS and Windows machines). \n\n# Beiboot Client\nThe Beiboot client is a Python API to manage Beiboot clusters. It can set up and tear down Kubernetes instances\nas well as establish connections.',
    'author': 'Michael Schilonka',
    'author_email': 'michael@blueshoe.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://getdeck.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import base64
import logging
import os
import pathlib
from typing import List, Optional

import docker.errors
import kubernetes as k8s
from beiboot.configuration import ClientConfiguration
from docker.models.containers import Container

logger = logging.getLogger("getdeck.beiboot")


def create_beiboot_custom_ressource(
    config: ClientConfiguration, name: str, ports: List[str]
) -> dict:
    cr = {
        "apiVersion": "getdeck.dev/v1",
        "kind": "beiboot",
        "provider": "k3s",
        "ports": ports,
        "metadata": {
            "name": name,
            "namespace": config.NAMESPACE,
        },
    }
    return cr


def decode_kubeconfig(kubeconfig_obj: dict):
    b64_kubeconfig = kubeconfig_obj["source"]
    kubeconfig = base64.b64decode(b64_kubeconfig.encode("utf-8")).decode("utf-8")
    return kubeconfig


def get_kubeconfig_location(config: ClientConfiguration, cluster_name: str) -> str:
    return config.KUBECONFIG_LOCATION.joinpath(f"{cluster_name}.yaml")


def save_kubeconfig_to_file(
    config: ClientConfiguration, cluster_name: str, kubeconfig
) -> str:
    location = get_kubeconfig_location(config, cluster_name)
    with open(location, "w") as yaml_file:
        yaml_file.write(kubeconfig)
    return location


def delete_kubeconfig_file(config: ClientConfiguration, cluster_name: str):
    location = get_kubeconfig_location(config, cluster_name)
    pathlib.Path(location).unlink(missing_ok=True)


def _get_tooler_container_name(cluster_name: str):
    return f"getdeck-proxy-{cluster_name}"


def start_kubeapi_portforwarding(config: ClientConfiguration, cluster_name: str):
    bbt = config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
        namespace=config.NAMESPACE,
        name=cluster_name,
        group="getdeck.dev",
        plural="beiboots",
        version="v1",
    )
    forwarded_ports = bbt.get("ports")
    forwards = []
    for idx, port in enumerate(forwarded_ports):
        forwards.append(
            (
                port.split(":")[0],
                [
                    "kubectl",
                    "port-forward",
                    "-n",
                    bbt["beibootNamespace"],
                    f"svc/port-{port.split(':')[1]}",
                    port,
                    "--address='0.0.0.0'",
                ],
            )
        )

    forwards.append(
        (
            config.BEIBOOT_API_PORT,
            [
                "kubectl",
                "port-forward",
                "-n",
                bbt["beibootNamespace"],
                "svc/kubeapi",
                f"{config.BEIBOOT_API_PORT}:{config.BEIBOOT_API_PORT}",
                "--address='0.0.0.0'",
            ],
        )
    )
    if config.KUBECONFIG_FILE:
        kubeconfig_path = config.KUBECONFIG_FILE
    else:
        from kubernetes.config import kube_config

        kubeconfig_path = os.path.expanduser(kube_config.KUBE_CONFIG_DEFAULT_LOCATION)
    for forward in forwards:
        try:
            _cmd = ["/bin/sh", "-c", f"{' '.join(forward[1])}"]
            logger.debug(_cmd)
            container = config.DOCKER.containers.run(  # noqa
                image=config.TOOLER_IMAGE,
                name=f"{_get_tooler_container_name(cluster_name)}-{forward[0]}",
                command=_cmd,
                restart_policy={"Name": "unless-stopped"},
                remove=False,
                detach=True,
                ports={f"{forward[0]}/tcp": int(forward[0])},
                environment=["KUBECONFIG=/tmp/.kube/config"],
                volumes=[f"{kubeconfig_path}:/tmp/.kube/config"],
            )
        except docker.errors.APIError as e:
            if e.status_code == 409:
                try:
                    _cmd = ["/bin/sh", "-c", f"{' '.join(forward[1])}"]
                    # retry
                    config.DOCKER.containers.get(
                        f"{_get_tooler_container_name(cluster_name)}-{forward[0]}"
                    ).remove()
                    config.DOCKER.containers.run(  # noqa
                        image=config.TOOLER_IMAGE,
                        name=f"{_get_tooler_container_name(cluster_name)}-{forward[0]}",
                        command=_cmd,
                        restart_policy={"Name": "unless-stopped"},
                        remove=False,
                        detach=True,
                        ports={f"{forward[0]}/tcp": int(forward[0])},
                        environment=["KUBECONFIG=/tmp/.kube/config"],
                        volumes=[f"{kubeconfig_path}:/tmp/.kube/config"],
                    )
                except docker.errors.APIError as e:
                    raise RuntimeError(
                        "Finally failed to set up local proxy for the Kubernetes API"
                    )
            else:
                logger.error(e)
                raise RuntimeError(
                    "Could not create the local proxy for the Kubernetes API"
                )


def _list_containers_by_prefix(
    config: ClientConfiguration, prefix: str
) -> List[Optional[Container]]:
    containers = config.DOCKER.containers.list()
    result = []
    for container in containers:
        if container.name.startswith(prefix):
            result.append(container)
    return result


def kill_kubeapi_portforwarding(config: ClientConfiguration, cluster_name: str) -> None:
    try:
        config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
            namespace=config.NAMESPACE,
            name=cluster_name,
            group="getdeck.dev",
            plural="beiboots",
            version="v1",
        )
    except k8s.client.exceptions.ApiException:
        pass

    try:
        containers = _list_containers_by_prefix(
            config, _get_tooler_container_name(cluster_name)
        )
        for container in containers:
            try:
                container.kill()
            except:  # noqa
                pass
            try:
                container.remove()
            except:  # noqa
                pass
    except docker.errors.APIError as e:
        logger.warning(str(e))
